/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/


#include "dialog_client.h"
#include "win_util.h"

Client*
dialog_client_new(Wm *w, Window win, Client *trans)
{

   client *c = base_client_new(w, win); 
   c->type = dialog;
   
   c->reparent     = &dialog_client_reparent;
   c->move_resize  = &dialog_client_move_resize;
   c->hide         = &dialog_client_hide;
   
   c->configure    = &dialog_client_configure;
   c->button_press = &dialog_client_button_press;
   c->redraw       = &dialog_client_redraw;
   c->show         = &dialog_client_show;
   c->destroy      = &dialog_client_destroy;
   c->get_coverage = &dialog_client_get_coverage;
   c->trans = trans;

   return c;

}

void
dialog_client_get_coverage(Client *c, int *x, int *y, int *w, int *h)
{
   int frm_size = dialog_client_title_height(c);
   *x = c->x;
   *y = c->y - frm_size;
   *w = c->width;
   *h = c->height + frm_size;
}

void
dialog_client_move_resize(Client *c)
{
   int frm_size = dialog_client_title_height(c);
   base_client_move_resize(c);
   XResizeWindow(c->wm->dpy, c->window, c->width, c->height);
   XMoveResizeWindow(c->wm->dpy, c->frame, c->x, c->y - frm_size, c->width,
		     c->height + frm_size + WBW(c));
}

void
dialog_client_hide(Client *c)
{
  dbg("hiding dialog window\n");
  XLowerWindow(c->wm->dpy, c->frame);
}

int
dialog_client_title_height(Client *c)
{
   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
      return 0;
   return theme_get_frame_size(c->wm->mbtheme, FRAME_DIALOG );
}

void
dialog_client_show(Client *c)
{
  dbg("**** showing dialog client *****\n");
  client_set_state(c, NormalState); 
  XMapRaised(c->wm->dpy, c->frame);
  XMapSubwindows(c->wm->dpy, c->frame);
  
  if (client_want_focus(c))
    {
      XSetInputFocus(c->wm->dpy, c->window,
		     RevertToPointerRoot, CurrentTime);
      c->wm->focused_client = c;
    }
  c->mapped = True;
}

void
dialog_client_reparent(Client *c)
{
   MBtheme *t = c->wm->mbtheme; /* TOGO: replace with mbtheme  */
   XSetWindowAttributes attr;
   int frm_size = dialog_client_title_height(c);

   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG)
   {
      c->frame = c->title_frame = c->window;
      return;
   }
   
   attr.override_redirect = True;
   attr.border_pixel = t->wbc->xcol.pixel;
   attr.background_pixel = t->wbc->xcol.pixel;
   attr.event_mask = ChildMask|ButtonPressMask|ExposureMask;
   
   c->frame =
      XCreateWindow(c->wm->dpy, c->wm->root, 0, 0,
		    c->width, c->height + frm_size + WBW(c), 
		    WBW(c) ? WBW(c) : 1,
		    CopyFromParent, CopyFromParent, CopyFromParent,
		    CWOverrideRedirect|CWBorderPixel
		    |CWEventMask|CWBackPixel,
		    &attr);
   
   attr.background_pixel = t->wbc->xcol.pixel;
   
   c->title_frame =
      XCreateWindow(c->wm->dpy, c->frame, 0, 0, c->width, frm_size, 0,
		    CopyFromParent, CopyFromParent, CopyFromParent,
		    CWOverrideRedirect|CWBackPixel|CWEventMask, &attr);
   
   /* Adjust main parented windows propertys */
   XSetWindowBorderWidth(c->wm->dpy, c->window, 0);
   XAddToSaveSet(c->wm->dpy, c->window);
   XSelectInput(c->wm->dpy, c->window,
		ButtonPressMask|ColormapChangeMask|PropertyChangeMask);
   
   XReparentWindow(c->wm->dpy, c->window, c->frame, 0, frm_size + WBW(c));
}


void
dialog_client_configure(Client *c)
{
   int max_w = c->wm->dpy_width - 32;
   int max_h = c->wm->dpy_height - 32;

   dbg("dialog wants %i x %i, %i x %i", c->x, c->y, c->width, c->height);
   
   if (c->width > max_w) c->width   = max_w;
   if (c->height > max_h) c->height = max_h;

   /* only repos if at 0,0 or off screen  */
   if ( !c->x || (c->x + c->width) > c->wm->dpy_width || (c->x < 0))
      c->x = (c->wm->dpy_width  - c->width)/2;
   
   if ( !c->y || (c->y + c->height) > c->wm->dpy_height)
      c->y = (c->wm->dpy_height - c->height)/2;
}

void
dialog_client_redraw(Client *c, Bool use_cache)
{
   if (c->flags & CLIENT_TITLE_HIDDEN_FLAG) return;
   
   if (c->backing == (Pixmap)NULL)
      client_init_backing(c, c->width, theme_get_frame_size(c->wm->mbtheme,
							   FRAME_DIALOG ));
   
   theme_paint( c->wm->mbtheme, c, FRAME_DIALOG );
   theme_paint_button(c->wm->mbtheme,c, ACTION_CLOSE, INACTIVE, FRAME_DIALOG );
   
   XSetWindowBackgroundPixmap(c->wm->dpy, c->title_frame, c->backing);
   XClearWindow(c->wm->dpy, c->title_frame);
}

void
dialog_client_button_press(Client *c, XButtonEvent *e)
{
   switch (client_do_button_ops(c, FRAME_DIALOG, e))
   {
      case ACTION_CLOSE:
	 client_deliver_delete(c);
	 break;
      case -1: 		 /* Cancelled  */
	 break;
      case 0:
	 wu_drag(c);     /* Not on button */
	 break;
   }
}
 
void dialog_client_destroy(Client *c)
{
  /*   if (c->wm->main_client != NULL)
       c->wm->main_client->show(c->wm->main_client); */
   base_client_destroy(c);
}
