/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#ifndef _STRUCTS_H_
#define _STRUCTS_H_

#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#ifdef USE_XFT
#include <X11/Xft/Xft.h>
#endif

#define TRUE  1
#define FALSE 0

#define APPLY_GRAVITY 1
#define REMOVE_GRAVITY -1

#define DOCK_HEIGHT 16
#define DOCK_MIN_HEIGHT 4

#define STACK_REMOVE 1
#define STACK_ADD   -1

#define BUTTON_NEXT_PRESSED 1
#define BUTTON_PREV_PRESSED 2
#define BUTTON_CLOSE_PRESSED 3

#define MENU_FLAG         (1<<1)
#define SINGLE_FLAG       (1<<2) /* when only a single client */
#define DBL_CLICK_FLAG    (1<<3)
#define TITLE_HIDDEN_FLAG (1<<4)
#define DESKTOP_RAISED_FLAG (1<<5)


#define CLIENT_FULLSCREEN_FLAG (1<<1)
#define CLIENT_TITLE_HIDDEN_FLAG (1<<2)

#define MAX_FRAMES 5
#define MAX_COLORS 5
#define MAX_FONTS 5
#define MAX_PIXMAPS 12  
#define MAX_LAYERS  10  /* per frame  */
#define MAX_BUTTONS 10  /* per frame  */

#define ACTIVE   1
#define INACTIVE 2

#define MB_CMD_SET_THEME 1
#define MB_CMD_EXIT      2
#define MB_CMD_DESKTOP   3

#define Wm wm       /* tmp to avoid breakage */
#define Client client

#define START_CLIENT_LOOP(w,c) (c) = (w)->head_client; do {
#define END_CLIENT_LOOP(w,c)   } while (((c) = (c)->next) \
				        && ((c) != (w)->head_client) );

#define WBW(c) ((c)->wm->mbtheme->wbw)

#ifdef DEBUG
#define dbg(txt, args... ) fprintf(stderr, "DEBUG: " txt "\n", ##args )
#else
#define dbg(txt, args... ) /* nothing */
#endif

#define MAX(x,y) ((x>y)?(x):(y))

typedef enum { dialog,
	       toolbar,
	       detached, 
	       override,
	       dock,
	       menu,
	       mainwin,
	       desktop  } Client_type;

typedef struct _client
{
#ifdef USE_XFT
   unsigned char *name;
#else
   char	*name;
#endif

   XSizeHints	*size;
   Window	window, frame, title_frame;
   struct _client *trans;
   Colormap	cmap;
   int		x, y, width, height;
   int		ignore_unmap;
   Pixmap       icon, icon_mask;

   Window button_wins[MAX_BUTTONS];
   
   int flags;
   
   XID          win_group;
   
   Client_type type;
   
#ifdef USE_XFT
    XftDraw     *xftdraw;
#endif

   void (* reparent)( struct _client* c );
   void (* redraw)( struct _client* c, Bool use_cache );
   void (* button_press) (struct _client *c, XButtonEvent *e);
   void (* move_resize)( struct _client* c );
   void (* configure)( struct _client* c );
   void (* get_coverage)( struct _client* c, int* x, int* y, int* h, int* w );
   void (* hide)( struct _client* c );
   void (* show)( struct _client* c );
   void (* iconize)( struct _client* c );
   void (* destroy)( struct _client* c );

   Bool mapped;
   
   struct _client *next;
   struct _client *prev;

   Pixmap backing;
   Bool have_cache;
   
   void* xtra;
   struct _wm *wm;
   
} client;

/*
typedef struct _pxm 
{
   Pixmap pixmap;
   Pixmap mask;
   int width;
   int height;
   
} Pxm ;


typedef struct _theme 
{
   XFontStruct* font;
   XFontStruct* toolbar_font;
   
#ifdef USE_XFT
   XftFont *xftfont;
   XftFont *toolbar_xftfont;
   XftColor xft_fg;
#endif

   Pxm buttons[6];
   Pxm tile;
   
   int win_border_width;
   int padding;
   Bool bevel;
   
   XColor bg_col;
   XColor text_col;
   XColor fg_col;  
   XColor fg_start_col, fg_end_col;

   GC text_gc, toolbar_text_gc, bg_gc, fg_gc, effects_gc, mask_gc, invert_gc;
   
   Cursor move_curs, resize_curs;
   
} Theme;
*/

typedef struct _wm_config
{
   Bool use_title;
   
   unsigned int modifier;
   
   char *theme_file;
   char display_name[128];

   Time dbl_click_time;

   int use_icons;

} Wm_config;

typedef struct _wm
{
   Display*     dpy;
   Atom         atoms[30];
   Window       root;
   int          screen;
   
   client*      head_client;    /* For ptr to beginning of list   */
   client*      focused_client; /* currently focused client       */
   client*      main_client;    /* currently viewable main client */
   XFontStruct* font;

   int dpy_width;
   int dpy_height;

   int lock; /* to go, replaced by flags */

   int flags;
   
   Window dockwin;
   int dockwin_offset;

   Cursor move_curs, resize_curs, curs;

   /* To do, set in an array */
   Atom wm_state, wm_change_state, wm_protos, wm_delete, wm_cmapwins,
      mb_theme, mb_command;

   /* Theme* theme; */
   struct _mbtheme* mbtheme;
   Wm_config *config;

   Window last_click_window;
   Time   last_click_time;
   Bool   next_click_is_not_double;

   Pixmap generic_icon, generic_icon_mask;

} wm;

/* New Theme structs */


typedef struct _geom_param {
   enum { pixel, percentage, object, textx, textw } unit;
   signed int value;
   signed int offset;
} Gparam;

typedef struct _tcolor {
   char *name;
   XColor xcol;
#ifdef USE_XFT
   XftColor xftcol;
#endif
} Tcolor;

typedef struct _tpixmap {
   char *name; /* need hash ? */
   int w;
   int h;
   Pixmap pxmp;
   Pixmap mask;
} Tpixmap;

typedef struct _tgradient {
   char *name;            /* need hash ? */
   Tcolor *start_col;
   Tcolor *end_col;
} Tgradient;

typedef struct _tfont {
#ifdef USE_XFT
   XftFont *font;
#else
   XFontStruct* font;
#endif
   char *name;

} Tfont;

typedef enum { left, center, right } Alignment;  

typedef struct _tlabel {
   Alignment justify;
   struct _tcolor *col;
   struct _tcolor *bgcol;
   struct _tfont  *font;
} Tlabel;


#define ACTION_CLOSE 1
#define ACTION_NEXT  2
#define ACTION_PREV  3
#define ACTION_MIN   4
#define ACTION_MAX   5
#define ACTION_MENU  6
#define ACTION_HIDE  7

typedef struct _tbutton {
   
   int action;

   Bool inputonly;
   Bool wants_dbl_click;
   
   Gparam *x;
   Gparam *y;
   Gparam *w;
   Gparam *h;

   Tpixmap *active;
   Tpixmap *inactive;
   
} Tbutton;


typedef enum { gradientHoriz, gradientVert, label,
	       plain, pixmap, pixmapTiled, picture, icon } layer_type;

typedef struct _tlayer {
   char *name;
   /* can reference by pointer structs in  */
   layer_type type;

   Gparam *x;
   Gparam *y;
   Gparam *w;
   Gparam *h;

   /* TODO: add union for this lot */
   Tcolor *color;

   Tpixmap *tpxmp;

   Tpixmap *picture;
   
   Tlabel *label;

   
} Tlayer;

#define FRAME_MAIN       1
#define FRAME_DIALOG     2
#define FRAME_UTILITY_MAX 3
#define FRAME_UTILITY_MIN 4
#define FRAME_MENU       5

typedef struct _tframe {
   int id;
   int size;
   struct _tlayer layers[MAX_LAYERS];
   int layer_cnt;
   struct _tbutton *buttons[MAX_BUTTONS];
   int button_cnt;
   struct _tcol *border_col;
   int label_w;
   int label_h;  /* needed ?  */
   int label_x;

   char *option_str;
   
   int cur_width;
   
} Tframe;

typedef struct _mbtheme {

   struct _tframe frames[MAX_FRAMES];
   int frame_cnt;
   
   struct _tpixmap *pixmaps[MAX_PIXMAPS];
   int pixmap_cnt;

   struct _tpixmap *pictures[MAX_PIXMAPS];
   int picture_cnt;
   
   Tcolor *colors[MAX_COLORS]; 
   int color_cnt;

   Tfont *fonts[MAX_FONTS]; 
   int font_cnt;

   GC gc;
   GC mask_gc;
   GC band_gc; 	/* for drag window  */

  /*   
#ifdef USE_XFT
   XftFont *font;
#else
   XFontStruct* font;
#endif
  */
   struct _wm *wm;

   int wbw; 	/* win border width */
   struct _tcolor *wbc;

   struct _tfont *default_font;
   struct _tcolor *default_font_color;

   
} MBtheme;

#define VERTICAL   1
#define HORIZONTAL 2
#define TMP_WIDTH 240


#endif
