/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "toolbar_client.h"

Client*
toolbar_client_new(Wm *w, Window win)
{
   client *c = base_client_new(w, win); 
   c->type = toolbar;
   
   c->configure    = &toolbar_client_configure;
   c->reparent     = &toolbar_client_reparent;
   c->button_press = &toolbar_client_button_press;
   c->redraw       = &toolbar_client_redraw;
   c->hide         = &toolbar_client_hide;
   c->iconize      = &toolbar_client_hide;
   c->show         = &toolbar_client_show;
   c->move_resize  = &toolbar_client_move_resize;
   c->get_coverage = &toolbar_client_get_coverage;
   c->destroy      = &toolbar_client_destroy;


   client_set_state(c,WithdrawnState); 	/* set initially to signal show() */
   
   return c;
}

void
toolbar_client_configure(Client *c)
{
   if (client_get_state(c) == IconicState) {
      ; 
   } else {
      c->y = c->wm->dpy_height - wm_total_toolbar_height(c)
	 - c->height - (2*WBW(c));
      c->x = toolbar_win_offset(c);
      c->width = c->wm->dpy_width - toolbar_win_offset(c) - (2*WBW(c));
   }
}

void
toolbar_client_move_resize(Client *c)
{

   int max_offset = theme_get_frame_size(c->wm->mbtheme,
					 FRAME_UTILITY_MAX );
   int min_offset = theme_get_frame_size(c->wm->mbtheme,
					 FRAME_UTILITY_MIN );
   int offset = toolbar_win_offset(c);
   base_client_move_resize(c);

   if (client_get_state(c) == NormalState)
   {
      XResizeWindow(c->wm->dpy, c->window, c->width, c->height);
      XMoveResizeWindow(c->wm->dpy, c->frame, c->x - max_offset,
		  c->y, c->width + max_offset, c->height );
      XMoveResizeWindow(c->wm->dpy, c->title_frame, 0,
		  0, max_offset , c->height );
   } else {
     if (min_offset)
       {
	 XMoveResizeWindow(c->wm->dpy, c->frame, c->x,
			   c->y, c->width + max_offset, offset );
	 XMoveResizeWindow(c->wm->dpy, c->title_frame, 0,
			   0, c->width + max_offset , min_offset );
       }
   }      
}

void
toolbar_client_reparent(Client *c)
{
    XSetWindowAttributes attr;
    MBtheme *t = c->wm->mbtheme;

    int frm_size = theme_get_frame_size(c->wm->mbtheme, FRAME_UTILITY_MAX );
   
    attr.override_redirect = True;
    attr.border_pixel      = t->wbc->xcol.pixel;
    attr.background_pixel  = t->wbc->xcol.pixel;;
    attr.event_mask        = ChildMask|ButtonPressMask|ExposureMask;
    
    c->frame =
       XCreateWindow(c->wm->dpy, c->wm->root, 0, c->y,
		     c->wm->dpy_width, c->height, WBW(c),
		     CopyFromParent, CopyFromParent, CopyFromParent,
		     CWOverrideRedirect|CWBorderPixel|CWEventMask|CWBackPixel,
		     &attr);
   
   attr.background_pixel = BlackPixel(c->wm->dpy, c->wm->screen);
   
   c->title_frame =
      XCreateWindow(c->wm->dpy, c->frame, 0, 0, frm_size, c->height, 0,
		    CopyFromParent, CopyFromParent, CopyFromParent,
		    CWOverrideRedirect|CWBackPixel|CWEventMask, &attr);

   XSetWindowBorderWidth(c->wm->dpy, c->window, 0);
   XAddToSaveSet(c->wm->dpy, c->window);
   XSelectInput(c->wm->dpy, c->window,
		ButtonPressMask|ColormapChangeMask|PropertyChangeMask);
   
   XReparentWindow(c->wm->dpy, c->window, c->frame,
		   toolbar_win_offset(c), 0);
}

void
toolbar_client_show(Client *c)
{
   int i;

   XFreePixmap(c->wm->dpy, c->backing);
   c->backing = (Pixmap)NULL;
   c->mapped = True;
   if (client_get_state(c) == WithdrawnState)
   {
      wm_restack(c->wm, c, - c->height - WBW(c));
   } else if (client_get_state(c) == IconicState) {
      wm_restack(c->wm, c, -(c->height - toolbar_win_offset(c)));
      c->y = c->y - ( c->height - toolbar_win_offset(c));
      c->x = theme_get_frame_size(c->wm->mbtheme,
					 FRAME_UTILITY_MAX );
   } else { return; /* were already shown */ }

   /* destroy any eisting buttons  */
   for (i=0; i<MAX_BUTTONS; i++)
      if (c->button_wins[i] != None)
      {
	 XDestroyWindow(c->wm->dpy, c->button_wins[i]);
	 c->button_wins[i] = None;
      }
   
   client_set_state(c,NormalState);
   toolbar_client_move_resize(c);
   
   XMapRaised(c->wm->dpy, c->frame);
   XMapRaised(c->wm->dpy, c->title_frame);
   XMapRaised(c->wm->dpy, c->window); 

   if (client_want_focus(c))
   {
      dbg("client wants focus");
      XSetInputFocus(c->wm->dpy, c->window,
		     RevertToPointerRoot, CurrentTime);
      c->wm->focused_client = c;
   } 

   base_client_show(c);

   /* make sure any dialog clients are raised above */
   if (c->wm->main_client) base_client_show(c->wm->main_client);
}

void
toolbar_client_hide(Client *c)
{
   int i;

   if (client_get_state(c) == IconicState) return;
   
   client_set_state(c,IconicState);

   c->ignore_unmap++;
   XUnmapWindow(c->wm->dpy, c->window);

   /* destroy any eisting buttons  */
   for (i=0; i<MAX_BUTTONS; i++)
      if (c->button_wins[i] != None)
      {
	 XDestroyWindow(c->wm->dpy, c->button_wins[i]);
	 c->button_wins[i] = None;
      }
   
   c->x=0;

   c->y = c->y + (c->height -/* toolbar_win_offset(c) */
		  theme_get_frame_size(c->wm->mbtheme,
				       FRAME_UTILITY_MIN )
		  );

   if (c->backing) XFreePixmap(c->wm->dpy, c->backing);

   c->backing = None;

   toolbar_client_move_resize(c);

   dbg("hiding toolbar y is now %i", c->y);

   wm_restack(c->wm, c, c->height - toolbar_win_offset(c));

}


void
toolbar_client_destroy(Client *c)
{
   dbg("toolbar got destroy");
   if (c->x == theme_get_frame_size(c->wm->mbtheme,
					 FRAME_UTILITY_MAX ))
   {
      wm_restack(c->wm, c, c->height+WBW(c));
   } else {
      wm_restack(c->wm, c, +theme_get_frame_size(c->wm->mbtheme,
					 FRAME_UTILITY_MIN )+WBW(c));
   }
   base_client_destroy(c);
}


void
toolbar_client_get_coverage(Client *c, int *x, int *y, int *w, int *h)
{
   *x = c->x; *y = c->y + WBW(c);
   *w = c->width + toolbar_win_offset(c);

   if (client_get_state(c) == NormalState)
   {
      *x = c->x - toolbar_win_offset(c);
      *h = c->height + WBW(c);
   } else {
      *h = toolbar_win_offset(c) + WBW(c);
   }
}


void
toolbar_client_button_press(Client *c, XButtonEvent *e)
{
   int frame_id;
   if (client_get_state(c) == NormalState)
      frame_id = FRAME_UTILITY_MAX;
   else
      frame_id = FRAME_UTILITY_MIN;
   
   switch (client_do_button_ops(c, frame_id, e))
   {
      case ACTION_CLOSE:
	 client_deliver_delete(c);
	 break;
      case ACTION_MIN:
	 toolbar_client_hide(c);
	 break;
      case ACTION_MAX:
	 toolbar_client_show(c);
	 break;
      case -1: 		 /* Cancelled  */
	 break;
      case 0:
	 /* Not on button */
	 break;
   }

}

int
toolbar_win_offset(Client *c)
{
   if (client_get_state(c) == IconicState)
   {
      return theme_get_frame_size(c->wm->mbtheme, FRAME_UTILITY_MIN );
   } else {
      return theme_get_frame_size(c->wm->mbtheme, FRAME_UTILITY_MAX );
   }

}

void
toolbar_client_redraw(Client *c, Bool use_cache)
{
   int max_offset = theme_get_frame_size(c->wm->mbtheme, FRAME_UTILITY_MAX );
   int min_offset = theme_get_frame_size(c->wm->mbtheme, FRAME_UTILITY_MIN );
   int tmp;
   Tframe *frm;

   if (use_cache && c->backing != None) return;
   if (client_get_state(c) == IconicState)
   {
     if (!min_offset) return;

      if (c->backing == None)
	 client_init_backing(c, c->width + max_offset, min_offset);

      /* HACK */
      tmp = c->width;
      c->width = c->width + max_offset;
      theme_paint( c->wm->mbtheme, c, FRAME_UTILITY_MIN);
      theme_paint_button(c->wm->mbtheme,c, ACTION_CLOSE,
			 INACTIVE, FRAME_UTILITY_MIN);
      theme_paint_button(c->wm->mbtheme,c, ACTION_MAX,
			 INACTIVE, FRAME_UTILITY_MIN);

      c->width = tmp;
   } else {
      int tmp2;

     if (!max_offset) return;

      if (c->backing == None)
	 client_init_backing(c, max_offset, c->height);

      /* HACK */
      frm = theme_get_frame(c->wm->mbtheme, FRAME_UTILITY_MAX );
      tmp = c->width;
      tmp2 = frm->size;
      c->width = max_offset;
      frm->size = c->height;
      theme_paint( c->wm->mbtheme, c, FRAME_UTILITY_MAX);
      theme_paint_button(c->wm->mbtheme,c, ACTION_MIN,
			 INACTIVE, FRAME_UTILITY_MAX);
      theme_paint_button(c->wm->mbtheme,c, ACTION_CLOSE,
			 INACTIVE, FRAME_UTILITY_MAX);

      c->width = tmp; frm->size = tmp2;
   }

   XSetWindowBackgroundPixmap(c->wm->dpy, c->title_frame, c->backing);
   XClearWindow(c->wm->dpy, c->title_frame);
   
}





