/* matchbox - a lightweight window manager

   Copyright 2002 Matthew Allum

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include "wm.h"
#include "config.h"

wm*
wm_new(int argc, char **argv)
{
   int *j;
   Cursor curs;   
   XSetWindowAttributes sattr; /* for root win */
   XpmAttributes xpm_attr;
   
   Wm *w = NULL;

   int keys_to_grab[] = { XK_Return, XK_Tab, XK_l, XK_r,
			  XK_p, XK_n, XK_Left, XK_Right,
			  XK_Up, XK_Down, XK_L, XK_R, 0 };
   
   if ( (w = malloc(sizeof(Wm))) == NULL) err("err out of memory");
   
   wm_load_config(w, &argc, argv);
   
   if ((w->dpy = XOpenDisplay(w->config->display_name)) == NULL) {
      err("can't open display! check your DISPLAY variable.");
      exit(1);
   }
   
   XSetErrorHandler(handle_xerror); 

   w->screen = DefaultScreen(w->dpy);
   w->root   = RootWindow(w->dpy, w->screen);

   xpm_attr.valuemask = 0;

   if (w->config->use_icons)
   {
      if (XpmReadFileToPixmap( w->dpy, w->root, GENERIC_ICON,
			       &(w->generic_icon), &(w->generic_icon_mask),
			       &xpm_attr )
	  != XpmSuccess )
      {
	 fprintf(stderr,"Failed to load %s . Disabling icons.", GENERIC_ICON); 
	 w->config->use_icons = 0;
      }
   }
   
   w->dpy_width  = DisplayWidth(w->dpy, w->screen);
   w->dpy_height = DisplayHeight(w->dpy, w->screen);
   w->flags = 0;

   w->next_click_is_not_double = True;
   
   w->wm_state = XInternAtom(w->dpy, "WM_STATE", False);
   w->wm_change_state = XInternAtom(w->dpy, "WM_CHANGE_STATE", False);
   w->wm_protos   = XInternAtom(w->dpy, "WM_PROTOCOLS", False);
   w->wm_delete   = XInternAtom(w->dpy, "WM_DELETE_WINDOW", False);
   w->wm_cmapwins = XInternAtom(w->dpy, "WM_COLORMAP_WINDOWS", False);
   w->mb_theme    = XInternAtom(w->dpy, "_MB_THEME", False);
   
   w->head_client    = NULL; /* A 'general' pointer for circular list */ 
   w->focused_client = NULL; /* Points to the currently focused client */
   w->main_client    = NULL; /* Points to the current 'main' client
				- eg the 'BIG' window - */

   ewmh_init(w);
   
   sattr.event_mask =  SubstructureRedirectMask
                       |SubstructureNotifyMask
                       |StructureNotifyMask;

   /* Tell root win we wanna be wm */
   XChangeWindowAttributes(w->dpy, w->root, CWEventMask, &sattr);

   /* grab keyboard */
   if (w->config->modifier != -1)
     for (j = keys_to_grab; *j; j++)
       XGrabKey(w->dpy, XKeysymToKeycode(w->dpy, *j), w->config->modifier,
		w->root, True, GrabModeAsync, GrabModeAsync);
   
    mbtheme_init(w, w->config->theme_file);

    curs = XCreateFontCursor(w->dpy, XC_right_ptr);
    XDefineCursor(w->dpy, w->root, curs);
    
    return w;
}

void
wm_usage(char *progname)
{
   printf("usage: %s %s [options ...]\n", progname, VERSION);
   printf("\t-display <string> \n");
   printf("\t-theme <string> \n");
   printf("\t-modifier <mod1|mod2|ctrl|none>\n");    
   printf("\t-use_titlebar <yes|no>\n");
   printf("\n");
   printf("Compile time options:\n");
#ifdef DEBUG
   printf("\tdebug build     yes\n");
#else
   printf("\tdebug build     no\n");
#endif

#ifdef USE_XFT
   printf("\tXFT support     yes\n");
#else
   printf("\tXFT support     no\n");
#endif

#ifndef STANDALONE
   printf("\tTheme support     yes\n");
#else
   printf("\tTheme support     no\n");
#endif

   printf("\n(c) 2002 Matthew Allum\n");
   

   exit(0);
}

void
wm_load_config(wm *w, int *argc, char *argv[])
{
   static XrmDatabase rDB, cmdlnDB;
   char *type;

   XrmValue value;
   
   static int opTableEntries = 4;
   static XrmOptionDescRec opTable[] = {
      {"-theme",       ".theme",           XrmoptionSepArg, (XPointer) NULL},
      {"-modifier",    ".modifier",        XrmoptionSepArg, (XPointer) NULL},
      {"-use_titlebar",".titlebar",        XrmoptionSepArg, (XPointer) NULL},
      {"-display",     ".display",         XrmoptionSepArg, (XPointer) NULL},
   };
   
   XrmInitialize();

   rDB = XrmGetFileDatabase(CONFDEFAULTS);   

   XrmParseCommand(&cmdlnDB, opTable, opTableEntries, "matchbox", argc, argv); 
   if (*argc != 1) wm_usage(argv[0]);

   /*
   if ((enviroment = getenv("XENVIROMENT")) == NULL)
   {
     enviroment = (char *)malloc(sizeof(char)*(strlen(getenv("HOME")+14)));
     strcpy(enviroment,getenv("HOME"));
     strcat(enviroment, "/.Xdefaults");
   }
   XrmCombineFileDatabase(enviroment, &rDB, True);
   */

   XrmCombineDatabase(cmdlnDB, &rDB, True);
   
   if ( (w->config = malloc(sizeof(Wm_config))) == NULL)
      err("err out of memory");

   /* defaults */
   w->config->use_title = True;
   w->config->modifier  = Mod2Mask;
   w->config->display_name[0] = '\0';
   w->config->dbl_click_time = 200;
   w->config->use_icons = 16;
   
   if (XrmGetResource(rDB, "matchbox.display",
		      "Matchbox.Display",
		      &type, &value) == True)
   {
      strncpy(w->config->display_name, value.addr, (int) value.size);
      w->config->display_name[value.size] = '\0';
   } else {
      if (getenv("DISPLAY") != NULL)
	 strcpy(w->config->display_name, (char *)getenv("DISPLAY"));
   }

   if (XrmGetResource(rDB, "matchbox.theme",
		      "Matchbox.Theme",
		      &type, &value) == True)
   {
#ifdef STANDALONE
     fprintf(stderr, "This matchbox build does not support themeing\n");
#else
     w->config->theme_file = (char *)malloc(sizeof(char)*(value.size+1));
     strncpy(w->config->theme_file, value.addr, (int) value.size);
     w->config->theme_file[value.size] = '\0';
     dbg("cmb line: got theme :%s ", w->config->theme_file);
#endif
   } else {
     w->config->theme_file = NULL;
   }
   
   if (XrmGetResource(rDB, "matchbox.titlebar", "Matchbox.Titlebar",
		      &type, &value) == True)
   {
      if(strncmp(value.addr, "no", (int) value.size) == 0)
      {
	 dbg("TURNING TITLE OFF");
	 w->config->use_title = False;
      }
   }

   if (XrmGetResource(rDB, "matchbox.modifier", "Matchbox.modifier",
		      &type, &value) == True)
   {
      if(strncmp(value.addr, "mod1", (int) value.size) == 0)
      {
	 w->config->modifier = Mod1Mask;
      } else if(strncmp(value.addr, "mod2", (int) value.size) == 0)
      {
	 w->config->modifier = Mod2Mask;
      } else if(strncmp(value.addr, "ctrl", (int) value.size) == 0)
      {
	 w->config->modifier = ControlMask;
      } else {
	 w->config->modifier = -1;
      }
   }
   
}

void
wm_init_existing(Wm *w)
{
   unsigned int nwins, i;
   Window dummyw1, dummyw2, *wins;
   XWindowAttributes attr;
   Client *c;
   
   XQueryTree(w->dpy, w->root, &dummyw1, &dummyw2, &wins, &nwins);
   for (i = 0; i < nwins; i++) {
      XGetWindowAttributes(w->dpy, wins[i], &attr);
      if (!attr.override_redirect && attr.map_state == IsViewable)
      {
	 c = wm_make_new_client(w, wins[i]);
	 c->ignore_unmap++; /* why do existing wins unmap - on reparent ? */
      }
   }
   XFree(wins);
}


Client*
wm_find_client(Wm *w, Window win, int mode)
{
    Client *c;
    if (w->head_client == NULL) return NULL;

    if (mode == FRAME) {
       START_CLIENT_LOOP(w,c);
       if (c->frame == win || c->title_frame == win) return c;
       END_CLIENT_LOOP(w,c);
    } else {
       START_CLIENT_LOOP(w,c);
       if (c->window == win) return c;
       END_CLIENT_LOOP(w,c);
    }    
    return NULL;
}


void
wm_event_loop(Wm* w)
{
    XEvent ev;
    for (;;) {
        XNextEvent(w->dpy, &ev);
        switch (ev.type) {
	   case ButtonPress:
	      dbg("event loop: button press");
	      wm_handle_button_event(w, &ev.xbutton); break;
	   case MapRequest:
	      dbg("event loop: map");
	      wm_handle_map_request(w, &ev.xmaprequest); break;
	   case UnmapNotify:
	      dbg("event loop: unmap");
	      wm_handle_unmap_event(w, &ev.xunmap); break;
	   case Expose:
	      dbg("event loop: expose");
	      wm_handle_expose_event(w, &ev.xexpose); break;
	   case DestroyNotify:
	      dbg("event loop: destroy");
	      wm_handle_destroy_event(w, &ev.xdestroywindow); break;
	   case ConfigureRequest:
	      dbg("event loop: conf request");
	      wm_handle_configure_request(w, &ev.xconfigurerequest); break;
	   case ConfigureNotify:
	      dbg("event loop: conf notify");
	      wm_handle_configure_notify(w, &ev.xconfigure); break;
	   case ClientMessage:
	      dbg("event loop: client message");
              wm_handle_client_message(w, &ev.xclient); break;
	   case KeyPress:
	      dbg("event loop: keypress");
	      wm_handle_keypress(w, &ev.xkey); break;
	   case PropertyNotify:
	      dbg("event loop: property notify");
	       wm_handle_property_change(w, &ev.xproperty); break;
	      /* TODO add these
            case ColormapNotify:
                handle_colormap_change(&ev.xcolormap); break;
	      */
	  default:
	     //dbg("unhandled event %d",ev.type);
	     break;
	}
    }
}

void
wm_handle_button_event(Wm *w, XButtonEvent *e)
{
   Client *p;
   Client *c = wm_find_client(w, e->window, FRAME);

   /* find out if this is a double click */
   if (w->next_click_is_not_double) {
      w->last_click_window = e->window;
      w->last_click_time   = e->time;
      w->next_click_is_not_double = False;
   } else {
      if (w->last_click_window == e->window
	  && ((e->time - w->last_click_time) < w->config->dbl_click_time))
      {
	 w->flags ^= DBL_CLICK_FLAG;
	 w->next_click_is_not_double = True;
      } else {
	 w->last_click_window = e->window;
	 w->last_click_time   = e->time;
	 w->next_click_is_not_double = False;
      }
   }

   /* remove menu if its up */
   if (w->flags & MENU_FLAG)
   {
      if (c && c->type == menu ) c->button_press(c,e);

      START_CLIENT_LOOP(w,p);
      if (p->type == menu)
      { select_client_destroy(p); }
      END_CLIENT_LOOP(w,p);
      return;
   }

   if (c) c->button_press(c,e);
   
   /* clear double click flag if set */
   if (w->flags & DBL_CLICK_FLAG) w->flags ^= DBL_CLICK_FLAG;
}

void
wm_handle_keypress(Wm *w, XKeyEvent *e)
{
   Client *p;
   KeySym key = XKeycodeToKeysym(w->dpy,e->keycode,0);
   if (w->config->modifier == -1) return;

   switch (key) {
      case XK_L:
	 XGrabServer(w->dpy);
	 XSync(w->dpy, False);	    
	 START_CLIENT_LOOP(w,p);
	 dbg("got L");
	 if (p->type == toolbar) p->hide(p);
	 END_CLIENT_LOOP(w,p);
	 XUngrabServer(w->dpy);
	 break;
      case XK_R:
	 XGrabServer(w->dpy);
	 XSync(w->dpy, False);	    
	 START_CLIENT_LOOP(w,p);
	 dbg("got R");
	 if (p->type == toolbar) p->show(p);
	 END_CLIENT_LOOP(w,p);
	 XUngrabServer(w->dpy);
	 break;
      case XK_l:
      case XK_Down:
	 XGrabServer(w->dpy);
	 XSync(w->dpy, False);	    
	 START_CLIENT_LOOP(w,p);
	 if (p->type == toolbar && p->window != w->dockwin) p->hide(p);
	 END_CLIENT_LOOP(w,p);
	 XUngrabServer(w->dpy);
	 break;
      case XK_r:
      case XK_Up:
	 XGrabServer(w->dpy);
	 XSync(w->dpy, False);	    
	 START_CLIENT_LOOP(w,p);
	 if (p->type == toolbar && p->window != w->dockwin) p->show(p);
	 END_CLIENT_LOOP(w,p);
	 XUngrabServer(w->dpy);
	 break;
      case XK_Return:
	    if (w->main_client)
	       client_deliver_delete(w->main_client);
	    break;
      case XK_Left:
      case XK_p:
	    if (w->main_client)
	    {
	       XGrabServer(w->dpy);
	       XSync(w->dpy, False);	    
	       p = client_get_prev(w->main_client, mainwin);
	       w->main_client->hide(w->main_client);
	       p->show(p);
	       XUngrabServer(w->dpy);
	    }
	 break;
      case XK_Tab:
      case XK_n:
      case XK_Right:
	 if (w->main_client)
	 {
	    XGrabServer(w->dpy);
	    XSync(w->dpy, False);	    
	    p = client_get_next(w->main_client, mainwin);
	    w->main_client->hide(w->main_client);
	    p->show(p);
	    XUngrabServer(w->dpy);
	 }
	 break;
   }
}

void
wm_handle_configure_notify(Wm *w, XConfigureEvent *e)
{
   Client *p;
   int height_diff, width_diff, b;
   
   if (e->window == w->root) /* screen rotation */
   {
      dbg("configure notify event called on root");
      if (e->width  != w->dpy_width ||
	  e->height != w->dpy_height)
      {
	 int i;
	 height_diff = e->height - w->dpy_height;
	 width_diff  = e->width  - w->dpy_width;
	 w->dpy_width = e->width; 
	 w->dpy_height = e->height;
	 b = WBW(w->main_client);
	 XGrabServer(w->dpy);
	 
	 START_CLIENT_LOOP(w,p);
	 switch (p->type)
	 {
	    case mainwin :
	       p->width += width_diff;
	       p->height += height_diff;
	       p->have_cache = False;
	       break;
	    case toolbar :
	       p->width += width_diff;
	       p->y += height_diff;
	       break;
	    case dock :
	       p->width += width_diff;
	       p->y += height_diff;
	       break;
	    case dialog :
	      dialog_client_configure(p);
	      break;
	    case override :
	    case menu:
	    case detached:
	    case desktop:
	       p->width += width_diff;
	       p->height += height_diff;
	       break;
	 }
	 p->move_resize(p);
	 /* destroy buttons so they get reposioned - maybe improve */
	 for(i=0;i<MAX_BUTTONS;i++)
	    if (p->button_wins[i] != None)
	    {
	       XDestroyWindow(w->dpy, p->button_wins[i]);
	       p->button_wins[i] = None;
	    }
	 
	 p->redraw(p, False);
	 client_deliver_config(p);
	 END_CLIENT_LOOP(w,p);
	 
	 XSync(w->dpy, False);
	 XUngrabServer(w->dpy);
      }
   }
}

void
wm_handle_configure_request(Wm *w, XConfigureRequestEvent *e)
{
   /* this needs work - ideally no window should be allowed to
      resize itself, only iconise its self exception is the dock .
   */
   Client *c = wm_find_client(w, e->window, WINDOW);
   XWindowChanges xwc;
   if (!c) {

       dbg("couldn't find window for configure request");
       xwc.x = e->x;
       xwc.y = e->y;
       xwc.width = e->width;
       xwc.height = e->height;
       xwc.sibling = e->above;
       xwc.stack_mode = e->detail;
       XConfigureWindow(w->dpy, e->window, e->value_mask, &xwc);
       return;

   }
   
   dbg("configure request event called on win %s", c->name);
   
   if (c->type == dock)
   {
      int old_height = c->height;
      c->height = e->height ;
      XResizeWindow(c->wm->dpy, c->window, c->width, c->height);
      XMoveWindow(c->wm->dpy, c->window, c->x, c->y - (c->height-old_height));
      wm_restack(c->wm, c,  old_height - c->height  );
      c->y = c->y - (c->height - old_height);
      return;
   } 

   xwc.width  = c->width;
   xwc.height = c->height;
   xwc.x = c->x;
   xwc.y = c->y;
   xwc.border_width = WBW(c);
   xwc.sibling = e->above;
   xwc.stack_mode = e->detail;
   
   XConfigureWindow(w->dpy, e->window,
		    e->value_mask, &xwc);

   if (e->detail == Above) 
      c->show(c);

}

void
wm_handle_map_request(Wm *w, XMapRequestEvent *e)
{
   Client *c = wm_find_client(w, e->window, WINDOW);

   if (!c) {
      wm_make_new_client(w, e->window);
   } else {
      dbg("Honoring map request for %s\n", c->name);
      c->show(c);
   }
}

void
wm_handle_unmap_event(Wm *w, XUnmapEvent *e)
{
   XEvent ev;
   Client *c = wm_find_client(w, e->window, WINDOW);
   if (!c) return;

   dbg(" unmap on %s ", c->name);
   if (c->ignore_unmap)
      c->ignore_unmap--;
   else
   {
      if (!c->mapped) return;
      XGrabServer(w->dpy);
      if (XCheckTypedWindowEvent(c->wm->dpy, c->frame, DestroyNotify,&ev)) {
	 dbg("Umap looks like destrroy on its way....");
	 wm_handle_destroy_event(w, &ev.xdestroywindow);
      } else {
	dbg("calling destroy on window\n");
	client_set_state(c, WithdrawnState);
	c->destroy(c);
	//XSync(c->wm->dpy, False);   
      }
      XUngrabServer(w->dpy);
   }
}


void
wm_handle_expose_event(Wm *w, XExposeEvent *e)
{
   Client *c = wm_find_client(w, e->window, FRAME);
    
   if (c && e->count == 0)
   {
      dbg("honoring expose event on %s", c->name);
      c->redraw(c, TRUE); /* redraw title from cache */
   }
}

void
wm_handle_destroy_event(Wm *w, XDestroyWindowEvent *e)
{
    Client *c = wm_find_client(w, e->window, WINDOW);
    if (!c) return; 
    dbg("destroy event called on %s", c->name);
    wm_remove_client(w, c);
}

void
wm_handle_client_message(Wm *w, XClientMessageEvent *e)
{
   Client *c = wm_find_client(w, e->window, WINDOW);

   dbg("client message type is %li\n" , e->message_type);

   if (e->message_type == w->atoms[MB_COMMAND])
     {				/* Handle messages from mbcontrol */
       switch (e->data.l[0])
	 {
#ifndef STANDALONE
	 case MB_CMD_SET_THEME :
	   {
	     Atom realType;
	     unsigned long n;
	     unsigned long extra;
	     int format;
	     int status;
	     char * value;
	     struct stat stat_info;
	     
	     status = XGetWindowProperty(w->dpy, w->root,
					 w->mb_theme, 0L, 512L, False,
					 AnyPropertyType, &realType,
					 &format, &n, &extra,
					 (unsigned char **) &value);
	     
	     if (status != Success || value == 0
		 || *value == 0 || n == 0)
	       {
		 dbg("no _MB_THEME set on root window");
	       } else {
		 dbg("switching theme to %s", value);
		 if (stat(value, &stat_info) == -1)
		   {
		     fprintf(stderr, "Matchbox: unable to open theme %s\n",
			     value);
		   } else {
		     dbg("switching theme to %s", value);
		     mbtheme_switch(w, value);
		   }
	       }
	     return;
	   }
#endif
	 case MB_CMD_EXIT      :
	   exit(0);
	 case MB_CMD_DESKTOP   :
	   wm_toggle_desktop(w);
	 }
       return;
     }
	
   if (e->message_type == w->wm_change_state)
     {
       if (c && e->format == 32 && e->data.l[0] == IconicState)
	 {
	   c->iconize(c);
	 }
       return;
     }

   ewhm_handle_root_message(w, e);   

}

void
wm_handle_property_change(Wm *w, XPropertyEvent *e)
{
   Client *c = wm_find_client(w, e->window, WINDOW);

   if (!c) return;
   dbg("property change on %s %li\n", c->name, e->atom );
   
   switch (e->atom) {  /* Catch name change */
      case XA_WM_NAME:
	 if (c->name) XFree(c->name);
	 XFetchName(w->dpy, c->window, &c->name);
	 c->redraw(c, FALSE);
	 break;
   }

   if (e->atom == w->wm_change_state)
     {
       if(client_get_state(c) == WithdrawnState)
	 c->destroy(c);
     }

}

Client*
wm_make_new_client(Wm *w, Window win)
{
   Window trans_win;
   Atom prop, realType;
   unsigned long n;
   unsigned long extra;
   int format;
   int status;
   Atom * value = NULL;
   client *c = NULL, *t = NULL;
   XSetWindowAttributes  attribs;
   XWMHints *wmhints;
   
   XGrabServer(w->dpy);

   dbg("got map request for new client");
   
   /* find out what type of window it wants to be ... */
   prop = XInternAtom(w->dpy, "_NET_WM_WINDOW_TYPE", True);
   if (prop != None)
   {
      status = XGetWindowProperty(w->dpy, win, prop, 0L, 1000000L,
				  0, XA_ATOM, &realType, &format,
				  &n, &extra, (unsigned char **) &value);
      if (status == Success)
      {
	 if (realType == XA_ATOM && format == 32)
	 {
	    dbg("got type atom\n");
	    if (value[0] == w->atoms[WINDOW_TYPE_DOCK])
	    {
	       dbg("got dock atom\n");
	       //c = docked_client_new(w, win);
	       c = dockbar_client_new(w, win);
	    }
	    if (value[0] == w->atoms[WINDOW_TYPE_TOOLBAR])
	    {
	       dbg("got toolbar atom\n");
	       c = toolbar_client_new(w, win);
	    }
	    if (value[0] == w->atoms[WINDOW_TYPE_DESKTOP])
	    {
	       dbg("got toolbar atom\n");
	       c = desktop_client_new(w, win);
	    }
	    if (value[0] == w->atoms[WINDOW_TYPE_SPLASH])
	    {
	       dbg("got splash atom\n");
	       c = dialog_client_new(w, win, NULL);
	       c->flags ^= CLIENT_TITLE_HIDDEN_FLAG;
	    }

	 } 
      }
   }

   /* check for transient */
   XGetTransientForHint(w->dpy, win, &trans_win);
   
   if ( trans_win && (trans_win != win))
   {
      dbg("new window has a transient");
      t = wm_find_client(w, trans_win, WINDOW);

      if (t == NULL) /* Its transient for root, so we use group ids to find
			a possible parent                                  */
      {
	 Client *p;
	 if ((wmhints = XGetWMHints(w->dpy, win)) != NULL)
	 {
	    if (wmhints->window_group && w->head_client != NULL)
	    {
	       START_CLIENT_LOOP(w,p);
	       if (wmhints->window_group == p->window)
	       { t = p; break; }
	       END_CLIENT_LOOP(w,p);
	    }
	 }
      }
      dbg("attempting to birth new dialog");
      c = dialog_client_new(w, win, t); /* if t is is NULL, dialog will
					      always be visible */
   }
   
   if (c == NULL) /* default to a main client */
   {
      /* make sure fullscreen window goes below any utility wins / docks etc */
      if (w->main_client && (w->main_client->flags & CLIENT_FULLSCREEN_FLAG))
	 main_client_hide(w->main_client);
      c = main_client_new(w, win);
   }

   if ((wmhints = XGetWMHints(w->dpy, win)) != NULL)
   {
      if (w->config->use_icons)
      {
	 if (wmhints->flags & IconPixmapHint)
	 {
	    dbg("wm_new: got icon hint"); 
	    c->icon      = wmhints->icon_pixmap;
	    if (wmhints->flags & IconMaskHint)
	    {
	       c->icon_mask = wmhints->icon_mask;
	    }
	    scale_icon(c);
	 } else {
	    /* set default */
	    c->icon = w->generic_icon;
	    c->icon_mask = w->generic_icon_mask;
	 }
      }
      XFree(wmhints);
   }
   
   
   dbg("calling configure for new client");
   
   c->configure(c);

   wm_gravitate(c, 1);

   dbg("reparenting new client");
   
   c->reparent(c);

   dbg("move/resizing  new client");
   
   c->move_resize(c);
   client_deliver_config(c);

   attribs.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
   attribs.save_under = False;
   XChangeWindowAttributes(w->dpy, c->window,
			   CWDontPropagate|CWSaveUnder,
			   &attribs);

   dbg("finally showing new client");
   c->show(c);

   ewmh_update(c->wm);

   XSync(w->dpy, False);
   XUngrabServer(w->dpy);
   return c;
}

void
wm_remove_client(Wm *w, Client *c)
{
    XGrabServer(c->wm->dpy);
    XSetErrorHandler(ignore_xerror);
    c->destroy(c);
    XSetErrorHandler(handle_xerror);
    XUngrabServer(w->dpy);
}

void
wm_gravitate(Client *c, int multiplier)
{
    int dy = 0;
    /*
    int gravity = (c->size->flags & PWinGravity) ?
        c->size->win_gravity : NorthWestGravity;

    switch (gravity) {
        case NorthWestGravity:
        case NorthEastGravity:
        case NorthGravity: 10 ; break;
         case CenterGravity: 10 ; break;
    }
    */
    c->y += multiplier * dy;
}

void
wm_restack(Wm *w, Client *c, signed int amount)
{
   client *p;
   XGrabServer(w->dpy);
   for (p=c->next; p != c; p=p->next)
   {
     dbg("restacking, comparing %i is less than %i for %s",
	 p->y, c->y, p->name);
      if (p->y <= c->y)
      {
	 dbg("restacking %s", p->name);
	 switch (p->type)
	 {
	    case mainwin :
	       if (p->flags & CLIENT_FULLSCREEN_FLAG) break;
	       p->height += amount;
	       p->move_resize(p);
	       client_deliver_config(p);
	       break;
	    case toolbar :
	       p->y += amount;
	       p->move_resize(p);
	       client_deliver_config(p);
	       break;
	    case dock :
	       p->y += amount;
	       p->move_resize(p);
	       client_deliver_config(p);
	       break;
	    case dialog :
	    case override :
	    case menu:
	 case detached:
	 case desktop:
	       break;
	 }
      }
   }
   XSync(w->dpy, False);
   XUngrabServer(w->dpy);
}

int
wm_total_toolbar_height(Client *c) /* find the toolbar height */
{
   client *p;
   int x,y,ww,h;
   int t = 0;

   START_CLIENT_LOOP(c->wm,p)
      if ( (p->type == toolbar || p->type == dock )
	    && c != p)
      {
	 p->get_coverage(p, &x, &y, &ww, &h);
	 t += h;
      }
   END_CLIENT_LOOP(c->wm,p)

   return t;
}

int
wm_total_dock_height(Client *c)
{
   client *p;
   int x,y,ww,h;
   int t = 0;

   START_CLIENT_LOOP(c->wm,p)
      if ( (p->type == dock ) && c != p)
      {
	 p->get_coverage(p, &x, &y, &ww, &h);
	 t += h;
      }
   END_CLIENT_LOOP(c->wm,p)

   return t;
}

void
wm_toggle_desktop(Wm *w)
{
   client *p;
   XGrabServer(w->dpy);
   if (w->flags & DESKTOP_RAISED_FLAG)
   {
      START_CLIENT_LOOP(w,p)
	 if ( p->type == mainwin)
	 {
	    p->show(p);
	 }
      END_CLIENT_LOOP(w,p);

      if (w->main_client)
	 w->main_client->show(p); 
   } else {
      START_CLIENT_LOOP(w,p)
	 if ( p->type == desktop) XMapRaised(w->dpy, p->window);
      END_CLIENT_LOOP(w,p);
      START_CLIENT_LOOP(w,p)
	 if ( p->type == toolbar) p->hide(p);
         if ( p->type == dock) 
             XMapRaised(w->dpy, p->window);
      END_CLIENT_LOOP(w,p);
   }
   XUngrabServer(w->dpy);
   w->flags ^= DESKTOP_RAISED_FLAG;
   
}



