#include "xml.h"

static void node_start_cb( XMLParser *parser, char *tag, Params *attr);
static void node_end_cb( XMLParser *parser, char *tag );
static void node_cdata_cb( XMLParser *parser, char *cdata);
static char *parse(XMLParser *parser, char *doc);
static char* load_file(const char* filename);
static void _xml_parser_free(XMLNode *node);
     
Nlist *list_add_node(XMLNode *dest, XMLNode *new)
{
   Nlist *tmp;

   if (dest->kids == NULL)
   {
      dest->kids = (Nlist *)malloc(sizeof(Nlist));
      tmp = dest->kids;
   } else {
      for(tmp = dest->kids; tmp->next != NULL; tmp = tmp->next);
      tmp->next = (Nlist *)malloc(sizeof(Nlist));
      tmp = tmp->next;
   }
   tmp->next = NULL;
   tmp->data = new;
   return tmp;
}

XMLNode *new_xmlnode(char *name, Params *attr)
{
   XMLNode *n = (XMLNode *)malloc(sizeof(XMLNode));
   n->tag = (char *)malloc(sizeof(char)*(strlen(name)+1));
   strcpy(n->tag, name);

   if (attr != NULL)
   {
      n->attr  = (Params *)malloc(sizeof(*attr));
      memcpy((void *)n->attr, attr, sizeof(*attr));
   } else {
      n->attr = NULL;
   }
   n->parent = NULL;
   n->cdata = NULL;
   n->kids = NULL;
   return n;
}

static void node_start_cb( XMLParser *parser, char *tag, Params *attr)
{
   if (parser->root_node == NULL)
   {
      parser->root_node = new_xmlnode(tag, attr);
      parser->_current_node = parser->root_node;
   } else {
      Nlist *tmp;
      tmp = list_add_node(parser->_current_node, new_xmlnode(tag, attr));
      tmp->data->parent = parser->_current_node;
      parser->_current_node = tmp->data; 
   }
}

static void node_end_cb( XMLParser *parser, char *tag )
{
   parser->_current_node = parser->_current_node->parent; 
}

static void node_cdata_cb( XMLParser *parser, char *cdata)
{
   /* current_node->cdata */
   if (cdata == NULL) return;
   if (parser->_current_node->cdata == NULL)
   {
      parser->_current_node->cdata =
	 (char *)malloc(sizeof(char)*(strlen(cdata)+1));
      strcpy(parser->_current_node->cdata, cdata);
   } else {
      parser->_current_node->cdata =
	 (char *)realloc(parser->_current_node->cdata,
			 sizeof(char)*(strlen(cdata)+1));
      strcat(parser->_current_node->cdata, cdata);
   }
}


void xml_dump(XMLNode *node, int depth)
{
   Nlist *tmp;
   Params *p;
   char crap[] = "-------";
   char more_crap[] = "       ";
   printf("+%s %s\n", &crap[5-depth], node->tag);
   
   for(p = node->attr; p != NULL; p = p->next)
   {
      printf(" %s %s = %s\n", &more_crap[5-depth], p->key, p->value);
   }

   if (node->cdata)
      printf(" %s%s\n", &more_crap[5-depth], node->cdata);
   
   for(tmp = node->kids; tmp != NULL; tmp = tmp->next)
   {
      xml_dump(tmp->data, depth+2);
   }
}

static
void _xml_parser_free(XMLNode *node)
{
   Nlist *tmp;

   /*
   for(p1 = node->attr; p2 != NULL; p2 = p2->next)
      if (p) free(p);

   if (node->cdata)
      free(node->cdata);
   */
   
   for(tmp = node->kids; tmp != NULL; tmp = tmp->next)
      _xml_parser_free(tmp->data);

   //free(node);
}



/* --------------------------------------------------------------- */

Params *new_params(void)
{
   Params *p = (Params *)malloc(sizeof(Params));
   p->next = NULL;
   return p;
}

static char *parse(XMLParser *parser, char *doc)
{
   char *p = doc;
   
   /* Temporary data */
   char *oe_data = NULL;
   char *e_data = NULL;
   char *ce_data = NULL;
   Params *attr_params = NULL, *start_params = NULL;
   int skip_spaces = 0;

   /* Find beginnin of first tag */
   while (*p && *p != '<') { p++; }

   /* Skip directives and comments */
   if (*(p+1) == '?') { p++; while (*p && *p != '<') { p++; } }
   if (*(p+1) == '!') { p++; while (*p && *p != '<') { p++; } }
   
   if (!*p) return NULL;
   
   oe_data = ++p;  /* Pointer to open tag name  */
   
   while (*p && *p != '>') /* advance to end of first tag  */
   {
      if (isspace(*p) && !skip_spaces)  /* possible attribute key */
      {
	 *p = '\0'; 
	 if (*(p+1) && !isspace(*(p+1))
	     && *(p+1) != '>' && *(p+1) != '/')
	 {
	    if (attr_params == NULL)
	    {
	       attr_params = new_params();
	       start_params = attr_params;
	    } else {
	       attr_params->next = new_params();
	       attr_params = attr_params->next;
	    }
	    attr_params->key = p+1;
	 }
	 p++;
      }
      else if (*p == '=')   /* beginning of attribute value  */
      {
	 *p = '\0'; 
	 p += 2; /* skipp first " */
	 attr_params->value = p;
	 skip_spaces = 1;      /* ignore spaces in value */
      }
      else if (*p == '"')      /* end of attribute value */
      {
	 *p = '\0'; 
	 skip_spaces = 0;
	 p++;
      } else {
	 p++;
      }
   }

   *p = '\0';    /* end possible dangling value pointer */

   parser->start_element_cb(parser, oe_data, start_params); /* fire all callback */
   
   
   if (*(p-1) == '/') {    /* catch <tag/>  */
      *(p-1) = '\0';
      parser->data_cb(parser, NULL);
      parser->end_element_cb(parser, oe_data);
      return p+1;          /* TODO: catch possible segfualt */
   }

   do {
      if (*p != '<') p++;  /* TODO: fixme ... maybe  */

      e_data = p;    	   /* grab cdata  */
      while (*p && *p != '<') p++;

      if (*(p+1) == '!') {  /* skip comments */
	 *p = '\0'; p++;
	 while (*p && *p != '>') p++;
	 parser->data_cb(parser, e_data);
	 
      } else {
      
	 if (!*p) return NULL;  /* catch possible segfualt  */
	 
	 if (*(p+1) == '/')   /* close tag  */
	 {
	    *p = '\0';
	    parser->data_cb(parser, e_data); /* cdata callback  */
	    p+=2;                 /* skip /> */
	    ce_data = p;
	    while (*p && *p != '>') p++; 
	    *p = '\0';
	    parser->end_element_cb(parser, ce_data);
	    return p+1;          /* TODO catch possible segfualt  */
	 } else {
	    *p = '\0';
	    parser->data_cb(parser, e_data);
	    *p = '<';
	    p = parse(parser, p);  /* recurse into inner tags */
	 }
      }

   } while (*p);

   return p;
}

static char* load_file(const char* filename) {
	struct stat st;
	FILE* fp;
	char* str;
	int len;

	/* Get the file size. */
	if (stat(filename, &st)) return NULL;

	/* Open the file. */
	fp = fopen(filename, "rb");
	if (!(fp = fopen(filename, "rb"))) return NULL;

	/* Read in the file. */
	str = (char *)malloc(sizeof(char)*(st.st_size + 1));
	len = fread(str, 1, st.st_size, fp);
	if (len >= 0) str[len] = '\0';

	fclose(fp);

	return str;
}

XMLParser *new_xmlparser(void)
{
   XMLParser *parser;
   parser = (XMLParser *)malloc(sizeof(XMLParser));
   parser->start_element_cb = NULL;
   parser->end_element_cb   = NULL;
   parser->data_cb          = NULL;
   parser->root_node        = NULL;
   parser->_current_node    = NULL;

   return parser;
}

void free_xmlparser(XMLParser *parser, XMLNode *root)
{
   if (root) _xml_parser_free(root);
   free(parser);
}

XMLNode *parse_data_dom(XMLParser *parser, char *data)
{
   parser->start_element_cb = &node_start_cb;
   parser->end_element_cb   = &node_end_cb;
   parser->data_cb          = &node_cdata_cb;
   if (parse(parser, data) == NULL) return NULL;
   return parser->root_node;
}

XMLNode *parse_file_dom(XMLParser *parser, char *filename)
{
   char *data = load_file(filename);
   if (data != NULL) return parse_data_dom(parser, data);
   return NULL;
}

/*
int main(int argc, char *argv[])
{
   XMLNode *root_node;
   XMLParser *parser = new_xmlparser();
   root_node = parse_file_dom(parser, "test2.xml");
   xml_dump(root_node, 0);

}
*/
