#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/xpm.h>
#include <X11/Xresource.h>
#include <X11/extensions/shape.h>

#include "minivolume.xpm"
#include "tray.h"

// http://www.opensound.com/pguide/mixer.html

Display* dpy;
Window win, dock_win;
GC gc, bg_gc;
int mixerfd;

int handle_xerror(Display *dpy, XErrorEvent *e)
{
   char msg[255];
   XGetErrorText(dpy, e->error_code, msg, sizeof msg);
   printf("X error (%#lx): %s (opcode: %i)\n",
	  e->resourceid, msg, e->request_code);
   exit(1);
}

int
xcol_from_str(Display *display, XColor *col, const char *defstr)
{
  char *str;
  const char delim[] = ",:";
  char *token;
  XColor exact;

  if ((strchr(defstr, delim[0]) != NULL)
      || (strchr(defstr, delim[1]) != NULL) )
  {
     str = strdup(defstr);

     token = strsep (&str, delim); 
     col->red = ( atoi(token) * 65535 ) / 255; 
     token = strsep (&str, delim); 
     col->green = ( atoi(token) * 65535 ) / 255;
     token = strsep (&str, delim); 
     col->blue = ( atoi(token) * 65535 ) / 255;

     return XAllocColor(display,
			DefaultColormap(display, DefaultScreen(display)),
			col);
  } else {
     return XAllocNamedColor(display,
			     DefaultColormap(display, DefaultScreen(display)),
			     defstr, col, &exact);
  }
}

void draw_vol()
{
   int i;
   XFillRectangle(dpy, win, bg_gc, 0, 0, 33, 110);
   for(i=0;i<=50;i++)
      XDrawLine(dpy, win, gc, 4+((i/2)), 4+(i*2), 4+25, 4+(i*2));
}


void draw_slider(int vol)
{
   XDrawRectangle(dpy, win, gc, 2, (100-vol), 29, 8); 
}

int set_vol(int vol)
{
  int new_vol;
  if (vol>100) vol = 100;
  if (vol<0)   vol = 0;
  
  new_vol = vol | (vol << 8);
  if (ioctl(mixerfd, SOUND_MIXER_WRITE_VOLUME, &new_vol)
      == -1)
    printf("unable to set volume\n");
    
  draw_vol();
  draw_slider(vol);

  return vol;
}


void usage(char *prog_name)
{
   printf("Usage: minivol [options...]\n");
   printf("Options:\n");
   printf("  -display  <display>\n");
   printf("  -fg <color>  set foreground colour\n");
   printf("  -bg <color>  set background colour\n");
   printf("  -r  align right in  dock ( matchbox only )\n");
   printf("  -h  this help\n\n");
   exit(1);
}

int main( int   argc,
          char **argv )
{
   Atom window_type_atom;
   Atom window_type_dock_atom;
   Pixmap icon, mask;
   XpmAttributes xpm_attr;	

   Atom mb_align_atom;
   Atom mb_align_east_atom;
   int align_dock_right = False;
   
   XSetWindowAttributes attr;
   XWMHints wm_hints;
   
   XGCValues gv;
   XEvent an_event;
   
   int done = 0;
   int vol, orig_vol;

   int is_popped = 0;
   
   char display_name[128]; 
   char *name = "/dev/mixer";
   int devmask = 0;

   static XrmDatabase rDB, cmdlnDB;
   char *enviroment;
   char *type;
   char buffer[20];
   XrmValue value;

   static int opTableEntries = 4;
   static XrmOptionDescRec opTable[] = {
      {"-bg",       ".background", XrmoptionSepArg, (XPointer) NULL},
      {"-fg",       ".foreground", XrmoptionSepArg, (XPointer) NULL},
      {"-display",  ".display",    XrmoptionSepArg, (XPointer) NULL},
      {"-r",        ".dock_right", XrmoptionNoArg, (XPointer) "yes"},
   };
   
   XColor fg_col, bg_col;

   char **argv_copy;  /* make a copy of argc as XResources drinks it */
   int argc_copy;
   int i;
   
   argc_copy = argc;
   argv_copy = (char **)malloc(sizeof(char*)*argc);
   for(i=0; i<argc; i++)
   {
      argv_copy[i] = (char *)malloc(sizeof(char)*(strlen(argv[i])+1));
      strcpy(argv_copy[i], argv[i]);
   }

   
   if ((mixerfd = open(name, O_RDWR)) < 0) {
      perror(name);
      exit(1);
   }

   if (ioctl(mixerfd, SOUND_MIXER_READ_DEVMASK, &devmask) == -1) {
      perror("SOUND_MIXER_READ_DEVMASK");
      exit(-1);
   }

   /* SOUND_MIXER_WRITE to set volume ... */
   if (ioctl(mixerfd, SOUND_MIXER_READ_VOLUME, &orig_vol)==-1)
   {
      printf("cant open mixer :(\n");
      exit(1);
   }
   vol = ( (orig_vol & 0xff) + ((orig_vol >> 8) & 0xff) ) / 2;

   /* resources */

   
   XrmInitialize();

   if ((enviroment = getenv("XENVIROMENT")) == NULL)
   {
      enviroment = (char *)getenv("HOME");
      strcat(enviroment, "/.Xdefaults");
   }
   
   rDB = XrmGetFileDatabase(enviroment);   

   XrmParseCommand(&cmdlnDB, opTable, opTableEntries, "minivol", &argc, argv); 

   if (argc != 1) usage(argv[0]);

   XrmCombineDatabase(cmdlnDB, &rDB, True);

   if (XrmGetResource(rDB, "minivol.display",
		      "Minivol.Display",
		      &type, &value) == True)
   {
      strncpy(display_name, value.addr, (int) value.size);
      display_name[value.size] = '\0';
   } else {
      if (getenv("DISPLAY") != NULL)
	 strcpy(display_name, (char *)getenv("DISPLAY"));
   }

   if ((dpy = XOpenDisplay(display_name)) == NULL)
   {
      printf("Cant open display\n"); exit(1);
   }

   xcol_from_str(dpy, &fg_col, "black");
   xcol_from_str(dpy, &bg_col, "white");

   if (XrmGetResource(rDB, "minivol.foreground",
		      "Minivol.Foreground",
		      &type, &value) == True)
   {
      strncpy(buffer, value.addr, (int) value.size);
      buffer[value.size] = '\0';
      xcol_from_str(dpy, &fg_col, buffer);
   }

   if (XrmGetResource(rDB, "minivol.background",
		      "Minivol.Background",
		      &type, &value) == True)
   {
      strncpy(buffer, value.addr, (int) value.size);
      buffer[value.size] = '\0';
      xcol_from_str(dpy, &bg_col, buffer);
   }

   if (XrmGetResource(rDB, "minivol.dock_right",
		      "Minivol.Dock_right",
		      &type, &value) == True)
   {
      align_dock_right = True;
   }
   XSetErrorHandler(handle_xerror); 

   /* -- Dock Window --  */

   mb_align_atom = XInternAtom(dpy, "_MB_DOCK_ALIGN", False);
   mb_align_east_atom = XInternAtom(dpy, "_MB_DOCK_ALIGN_EAST",False);

   dock_win = XCreateSimpleWindow(dpy,
				  RootWindow(dpy, DefaultScreen(dpy)),
				  0, 0,
				  15, 13,
				  0, BlackPixel(dpy, DefaultScreen(dpy)),
				  WhitePixel(dpy, DefaultScreen(dpy)));

   gv.graphics_exposures = False;
   gv.function           = GXcopy;
   gv.foreground         = fg_col.pixel;
   gc = XCreateGC(dpy, dock_win,
		  GCGraphicsExposures|GCFunction|GCForeground, &gv);
   gv.foreground         = bg_col.pixel;
   bg_gc = XCreateGC(dpy, dock_win,
		     GCGraphicsExposures|GCFunction|GCForeground, &gv);

   xpm_attr.valuemask = XpmColormap | XpmCloseness;
   xpm_attr.closeness = 40000;
   xpm_attr.colormap = DefaultColormap(dpy, DefaultScreen(dpy));

   if (XpmCreatePixmapFromData( dpy, dock_win, minivolume_xpm,
			    &icon, &mask, &xpm_attr) != XpmSuccess )
   {
      fprintf(stderr, "failed to get icon image\n");
      exit(1);
   }
       
   XCopyArea(dpy, icon, dock_win, gc, 0, 0,
	     xpm_attr.width, xpm_attr.height, 0, 0);
   
   XShapeCombineMask (dpy, dock_win, ShapeBounding, 0, 0, mask, ShapeSet);
   
   
   if (align_dock_right)
      XChangeProperty(dpy, dock_win, mb_align_atom, XA_ATOM, 32, 
		      PropModeReplace,
		      (unsigned char *) &mb_align_east_atom, 1);
   
   XStoreName(dpy, dock_win, "Volume Control");

   tray_init_session_info(dpy, dock_win, argv_copy, argc_copy);
   tray_init(dpy, dock_win);

   
   XSelectInput(dpy, dock_win, ExposureMask|ButtonPressMask ); 

   /* -- Popup Window -- */
   
   attr.override_redirect = True;
   attr.event_mask = ButtonPressMask|ExposureMask;
   attr.background_pixel = bg_col.pixel;
   attr.border_pixel = fg_col.pixel;
      
   win = XCreateWindow(dpy, RootWindow(dpy, DefaultScreen(dpy)),
		       400, 100, 33, 110, 1,
		       DefaultDepth(dpy, DefaultScreen(dpy)),
		       CopyFromParent,
		       DefaultVisual(dpy, DefaultScreen(dpy)),
		       CWOverrideRedirect|CWEventMask|
		       CWBackPixel|CWBorderPixel,
		       &attr);

   wm_hints.input = False;
   wm_hints.flags = InputHint;
   XSetWMHints(dpy, win, &wm_hints );


   
   XSelectInput(dpy, win, 
		VisibilityChangeMask|ExposureMask|ButtonPressMask|
		ButtonReleaseMask|ButtonMotionMask);

   {
      int is_sliding = 0; int new_vol;
      while (!done)
      {
	 while ( XPending(dpy) ) 
	 {
	    XNextEvent(dpy, &an_event);
	    switch (an_event.type) {
	       case VisibilityNotify:
		  if (is_popped) XMapRaised(dpy, win);
		  break;
	       case ButtonPress:
		  if (an_event.xmotion.window == win)
		  {
		     if ( (an_event.xmotion.y >= ((100-vol)+2))
			  && ( an_event.xmotion.y <= ((100-vol)+8)) )
			is_sliding = 1;
		     else
		       orig_vol= vol = set_vol(100-(an_event.xmotion.y-2));
		  } else {
		     if (is_popped)
		     {
		       XUngrabPointer(dpy, CurrentTime);
		       XUnmapWindow(dpy, win);
		       is_popped = 0;
		     } else {
			/*
			XGetWindowAttributes(dpy,dock_win,&rattr);
			printf("x : %i, y : %i\n", rattr.x,rattr.y);
			*/

			Window mouse_root, mouse_win;
			int win_x, win_y, x, y;
			unsigned int mouse_mask;
		       XGrabPointer(dpy, 
				    RootWindow(dpy, DefaultScreen(dpy)), True,
				    (ButtonPressMask|ButtonReleaseMask),
				    GrabModeAsync,
				    GrabModeAsync, None, None, CurrentTime);

			XQueryPointer(dpy, RootWindow(dpy, DefaultScreen(dpy)),
				      &mouse_root, &mouse_win,
				      &x, &y, &win_x, &win_y, &mouse_mask);

			if (y < 118 ) y = 136; 
			
			XMoveWindow(dpy,win,x-8,y-118);
			XMapRaised(dpy, win);
			is_popped = 1;
		     }
		  }
		  break;
	       case ButtonRelease:
		  is_sliding = 0;
		  break;
	       case MotionNotify:
		  if (is_sliding)
		  {
		    int vol = 100-(an_event.xmotion.y-2);
		    if (abs(vol-orig_vol) > 5)
		      {
			orig_vol = set_vol(vol);
		       }
		  }
		  break;
	       case Expose:
		  if (an_event.xexpose.window == dock_win)
		  {
		     XCopyArea(dpy, icon, dock_win, gc, 0, 0,
			       xpm_attr.width, xpm_attr.height, 0, 0);
		  } else {
		     if (is_popped) {
			draw_vol();
			draw_slider(vol);
		     }
		  }
		  break;
	    }
	    tray_handle_event(dpy, dock_win, &an_event);
	 }
	 usleep(10000L); /* sleep for a 10th of a second */
      }
   }
	 
   XCloseDisplay(dpy);
   close(mixerfd);
   
   return(0);
}
