/*
 * xsingleinstance
 *
 * Program to ensure that a single instance of an X window program
 * exists on the screen.
 *
 * Copyright (c) 2000, Merle F. McClelland for CompanionLink
 * Portions of this code, such as TryChildren and ClientWindow, are
 * based upon routines in the Xmu library, which is Copyright 1989, 1998,
 * The Open Group.  All Rights Reserved.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * See the files COPYRIGHT and LICENSE for distribution information.
*/

#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <signal.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xlibint.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xproto.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <X11/extensions/shape.h>

#define FALSE	0
#define TRUE	1

#include "tray.h"

Atom wm_state_atom    = 0;

/*
 * Window_With_Name: routine to locate a window with a given name on a display.
 *                   If no window with the given name is found, 0 is returned.
 *                   If more than one window has the given name, the first
 *                   one found will be returned.  Only top and its subwindows
 *                   are looked at.  Normally, top should be the RootWindow.
 */
Window Window_With_Name(dpy, top, name)
     Display *dpy;
     Window top;
     char *name;
{
        Window *children, dummy;
        unsigned int nchildren;
        int i;
        Window w=0;
        char *window_name;

        if (XFetchName(dpy, top, &window_name) && !strcmp(window_name, name))
          return(top);

        if (!XQueryTree(dpy, top, &dummy, &dummy, &children, &nchildren))
          return(0);

        for (i=0; i<nchildren; i++) {
                w = Window_With_Name(dpy, children[i], name);
                if (w)
                  break;
        }
        if (children) XFree ((char *)children);
        return(w);
}

int Window_With_xid(dpy, top, xid)
     Display *dpy;
     Window top;
     Window xid;
{
        Window *children, dummy;
        unsigned int nchildren;
        int i;
        int w=0;

	if (top == xid) return True;
	
        if (!XQueryTree(dpy, top, &dummy, &dummy, &children, &nchildren))
          return(0);

        for (i=0; i<nchildren; i++) {
                w = Window_With_xid(dpy, children[i], xid);
                if (w)
                  break;
        }
        if (children) XFree ((char *)children);
        return(w);
}


static Window
TryChildren(Display *dpy, Window win, Atom WM_STATE)
{
	Window root, parent;
	Window *children;
	unsigned int nchildren;
	unsigned int i;
	Atom type = None;
	int format;
	unsigned long nitems, after;
	unsigned char *data;
	Window inf = 0;

	if (!XQueryTree(dpy, win, &root, &parent, &children, &nchildren))
		return 0;
	for (i = 0; !inf && (i < nchildren); i++) 
	{
		XGetWindowProperty(dpy, children[i], WM_STATE, 0, 0, False,
	                        AnyPropertyType, &type, &format, &nitems,
	                        &after, &data);
		if (type)
			inf = children[i];
	}
	for (i = 0; !inf && (i < nchildren); i++)
		inf = TryChildren(dpy, children[i], WM_STATE);
	if (children) XFree((char *)children);
	return inf;
}

Window
ClientWindow(Display *dpy, Window win)
{
	 Atom type = None;
	 int format;
	 unsigned long nitems, after;
	 unsigned char *data;
	 Window inf;

	 if (!wm_state_atom)
	     return win;
	 XGetWindowProperty(dpy, win, wm_state_atom, 0, 0,
			    False, AnyPropertyType,
	                    &type, &format, &nitems, &after, &data);
	 if (type)
	     return win;
	 inf = TryChildren(dpy, win, wm_state_atom);
	 if (!inf)
	     inf = win;
	 return inf;
}

//
// Find a window's state
//
int
wm_state(Display *display, Window w)
{
	Atom real_type;
	int real_format, state = 0;
	unsigned long n, extra;
	unsigned char *data;

	if (XGetWindowProperty(display, w, wm_state_atom, 0L, 2L, False, AnyPropertyType,
	   		&real_type, &real_format, &n, &extra, &data) == Success && n) 
	{
		state = *(int *)data;
		XFree(data);
		return state;
	}
	else
		return -1;
}

//
// bring a window to the foreground
//
void raise_window(Display *display, Window w)
{
	int iState;
	iState = wm_state(display, w);
	if (iState == NormalState)
	 		XMapRaised(display, w);
  	else if( iState == IconicState )
	 		XMapRaised(display, w);
}

void usage()
{
   fprintf(stderr, "usage: monolauch [options..] xpm program \n");
   fprintf(stderr, "where options are;\n");
   fprintf(stderr, "-d <display>\n");
   fprintf(stderr, "-n <window name> window name of launched program\n");
   fprintf(stderr, "-k kill app, rather than iconize\n");
   fprintf(stderr, "-l launch new clients, rather than map existing ones\n");
   fprintf(stderr, "-m message the dock with output of program \n");
   fprintf(stderr, "-t <title> set the dock window title\n");
   fprintf(stderr, "-r align in right of dock ( matchbox wm only )\n");
}



int main(int argc, char *argv[])
{
	// We just use the DISPLAY environment variable
	char *disp = NULL;
	char *shellToUse;
	int xid = 0;
	int debug;
	int kill_dont_icon = 0;
	int switch_count = 0;
	int just_keep_launching = 0;
	char *user_window_title = NULL;
	Bool send_message = False;
	int i, win_x = 0, win_y = 0;
	
	Display *display;
	int screen;
	Atom window_type_atom;
	Atom window_type_dock_atom;
	Window win;
	Pixmap icon, mask;
	GC gc;
	XGCValues *gv = NULL;
	XpmAttributes attrib;	
	XEvent an_event;
	XSizeHints size_hints;
	Atom wm_delete, wm_protos;
	
	Atom mb_align_atom;
	Atom mb_align_east_atom;
	int align_dock_right = False;
	int not_done = 1;
	char *win_name_to_find = NULL;
	int opt_cnt = 0;
	char window_title_str[256];

	
	debug = 0;


	for (i=1; argv[i]; i++) {
	   char *arg = argv[i];
	   if (*arg=='-') {
	      switch (arg[1]) {
		 case 'd' :
		    disp = argv[i+1];
		    i++; switch_count += 2;
		    break;
		 case 'k' :
		    kill_dont_icon++;
		    switch_count++;
		    break;
		 case 'r' :
		    align_dock_right = True;
		    switch_count++;
		    break;
		 case 'l' :
		    just_keep_launching = True;
		    switch_count++;
		    break;
		 case 'm' :
		    just_keep_launching = True;
		    send_message = True;
		    switch_count++;
		    break;
		 case 'w' :
		    i++;
		    if (argv[i])
		       win_name_to_find = argv[i];
		    switch_count += 2;
		    break;
		 case 't' :
		    i++;
		    if (argv[i])
		       user_window_title = argv[i];
		    switch_count += 2;
		    break;


		 default:
		    usage();
		    exit(0);
		    break;
	      }
	   } else {
	      opt_cnt++;
	      if (opt_cnt == 1) i=argc;
	      break;
	   }	
	}
	

	
	if ((argc-switch_count) < 3)
	{
	   usage();
	   exit(1);
	}

	// Determine which shell to use to fork processes

	shellToUse = getenv("SHELL");
	if (!shellToUse)
		shellToUse = "/bin/sh";

	if ((display = XOpenDisplay(disp)) == NULL)
	{
		fprintf(stderr, "Cannot connect to X server on display %s.",
			XDisplayName(disp));
		exit(1);
	}

	wm_delete = XInternAtom(display, "WM_DELETE_WINDOW", False);
	wm_protos = XInternAtom(display, "WM_PROTOCOLS", False);
	window_type_atom = XInternAtom(display, "_NET_WM_WINDOW_TYPE" , False);
	window_type_dock_atom = XInternAtom(display,
					    "_NET_WM_WINDOW_TYPE_DOCK",False);
	wm_state_atom = XInternAtom(display, "WM_STATE", True);
	mb_align_atom = XInternAtom(display, "_MB_DOCK_ALIGN", False);
	mb_align_east_atom = XInternAtom(display, "_MB_DOCK_ALIGN_EAST",False);

	screen = DefaultScreen(display);

	win = XCreateSimpleWindow(display,
				  RootWindow(display, screen),
				  0, 0,
				  32, 16,
				  0, BlackPixel(display, screen),
				  WhitePixel(display, screen));

	XChangeProperty(display, win, window_type_atom, XA_ATOM, 32, 
			PropModeReplace, (unsigned char *)
			&window_type_dock_atom, 1);

	
	gc = XCreateGC(display, RootWindow(display, screen), 0, gv);
	
	attrib.valuemask = XpmCloseness;
	attrib.closeness = 40000;

	if (XpmReadFileToPixmap( display, win, argv[1+switch_count],
				 &icon, &mask, &attrib) != XpmSuccess )
	{
	   fprintf(stderr, "monolauch: failed loading image '%s'\n",
		   argv[1+switch_count]);
	   exit(1);
	}

       XResizeWindow(display, win, attrib.width, attrib.height);

       size_hints.flags = PPosition | PSize | PMinSize;
       size_hints.x = 0;
       size_hints.y = 0;
       size_hints.width      =  attrib.width;
       size_hints.height     =  attrib.height;
       size_hints.min_width  =  attrib.width;
       size_hints.min_height =  attrib.height;
       
       XSetStandardProperties(display, win, "mono", 
			      "mono", 0,
			      argv, argc, &size_hints);
       if (user_window_title != NULL)
	 {
	   XStoreName(display, win, user_window_title);
	 }
       else
	 {
	   strcpy(window_title_str, argv[2+switch_count]);
	   strcat(window_title_str, " Launcher");
	   XStoreName(display, win, window_title_str);
	 }
       XCopyArea(display, icon, win, gc, 0, 0,
		 attrib.width, attrib.height, 0, 0);
       
       XShapeCombineMask (display, win, ShapeBounding, 0, 0, mask, ShapeSet);


       if (align_dock_right)
	  XChangeProperty(display, win, mb_align_atom, XA_ATOM, 32, 
			  PropModeReplace,
			  (unsigned char *) &mb_align_east_atom, 1);

       tray_init_session_info(display, win, argv, argc);
       tray_init(display, win);
       
       XSelectInput(display, DefaultRootWindow(display),
		    SubstructureNotifyMask);
       
       XSelectInput(display, win, 
		    StructureNotifyMask|ExposureMask|ButtonPressMask|ButtonReleaseMask ); 
       signal(SIGCHLD, SIG_IGN);
       while (1) {
	  XNextEvent(display, &an_event);
	  switch (an_event.type) 
	    {
	    case ConfigureNotify:
	      win_x = an_event.xconfigure.x;
	      win_y = an_event.xconfigure.y;
	      break;
	    case Expose:
	      
		XCopyArea(display, icon, win, gc, 0, 0,
			  attrib.width, attrib.height, 0, 0);
		XShapeCombineMask (display, win, ShapeBounding,
				   0, 0, mask, ShapeSet);
		break;

	     case ButtonPress:
	       /* XMoveWindow(display, win, win_x+1, win_y+1); */
	       break;
	     case ButtonRelease:
	       /* XMoveWindow(display, win, win_x-1, win_y-1); */
		if (win_name_to_find)
		{
		  xid = Window_With_Name(display, RootWindow(display, screen),
		  win_name_to_find);
		} else {
		   if (xid
		       && (!Window_With_xid(display,
					    RootWindow(display, screen), xid)
			   ))
		      xid = 0;
		}
		if (xid == 0)
		  {
		    if (!send_message) {
		      switch (fork())
			{
			case 0: // Child process
			  execvp(argv[2+switch_count],
				 &argv[2+switch_count]);
			  fprintf(stderr, "Can't exec \"%s\"\n",
				  argv[2+switch_count]);
			  exit(1);
			case -1:
			  fprintf(stderr, "Couldn't fork\n");
			  break;
			}
		    } else {
		      char buf[256];
		      char msg[1024] = "";
		      FILE *ptr = NULL;

		      if ((ptr = popen(argv[2+switch_count], "r")) != NULL)
			while (fgets(buf, 256, ptr) != NULL)
			  strcat(msg, buf);

		      tray_send_message(display, win, (unsigned char*) msg);

		    }
		   if (just_keep_launching) 
		     {
		       xid = 0; continue;
		     }
		       

		   if (!win_name_to_find)
		   {
		      not_done = 1;
		      while (not_done) {
			 XNextEvent(display, &an_event);
			 switch (an_event.type) {
			    case CreateNotify:
			       if (an_event.xcreatewindow.send_event) {
				  break;
			       }
			       if (an_event.xcreatewindow.override_redirect) {
				  break;
			       }
			       xid = an_event.xcreatewindow.window;
			       not_done--;
			 }
		      }
		   }

		    
		 } else {
		    XWindowAttributes xid_attrib;
		    XGetWindowAttributes(display, xid, &xid_attrib);
		    
		    if (xid_attrib.map_state == IsUnmapped ||
			wm_state(display, xid) != NormalState)
		    {
		       //printf("window is not normal / unmapped\n");
		       raise_window(display, xid);
		    }
		    else
		    {
		       if (kill_dont_icon)
		       {
			  int i, n, found = 0;
			  Atom *protocols;
			  XEvent e;
			  
			  if (XGetWMProtocols(display, xid,
					      &protocols, &n))
			  {
			     for (i=0; i<n; i++)
				if (protocols[i] == wm_delete) found++;
			     XFree(protocols);
			  }
			  
			  if (found)
			  {
			     e.type = ClientMessage;
			     e.xclient.window = xid;
			     e.xclient.message_type = wm_protos;
			     e.xclient.format = 32;
			     e.xclient.data.l[0] = wm_delete;
			     e.xclient.data.l[1] = CurrentTime;
			     XSendEvent(display, xid, False,
					NoEventMask, &e);
			  }
			  else
			  {
			     XKillClient(display, xid);
			  }
			  xid = 0;
		       } else {
			  XIconifyWindow(display, xid, screen);
		       }
		    }		 
		    break;
		 } 
	  }
	  tray_handle_event(display, win, &an_event);
	   
	}
       XCloseDisplay(display);
       exit(0);

}
